#!/usr/bin/env python3
"""
2011-9-27
    A module to handle vcf file http://www.1000genomes.org/node/101
    Need to add row-based iterator.

Example:
    vcfFile = VCFFile(path=self.inputFname)
    trio_col_index_data = self.findTrioIndex(vcfFile.sample_id2index, self.trio_isq_id_ls)
    for vcfRecord in vcfFile.parseIter():
        locus_id = vcfRecord.locus_id
        chromosome = vcfRecord.chromosome
        pos = vcfRecord.pos
        pos = int(pos)
    
"""
import os, sys
import re, copy, csv
from palos import utils
from palos.ProcessOptions import  ProcessOptions
from palos.utils import PassingData, getColName2IndexFromHeader

diploidGenotypePattern = re.compile(r'([\d.])[|/]([\d.])')
    #".|.", "0|1" for phased
    #"./.", "0/1" for unphased

def parseOneVCFRow(row, col_name2index, col_index_individual_name_ls,
    sample_id2index, minDepth=1,\
    dataEntryType=1):
    """
    2014.01.08 fix a bug that skips calls and shortens data_row. 
    2012.9.6 turn pos into integer
    2012.5.10
        complete representation of one locus
    2012.1.17
        common snippet split out of VCFFile & VCFRecord
        row is a list of input columns from one VCF file line
        dataEntryType
            1: each cell is base call
            2: each cell is a dictionary {'GT': base-call, 'DP': depth}
    """
    chromosome = row[0]
    pos = int(row[1])	#2012.9.6 turn pos into integer
    vcf_locus_id=row[2]
    quality = row[5]
    filter=row[6]
    info = row[7]
    format = row[8]
    info_ls = info.split(';')
    info_tag2value = {}
    for info_entry in info_ls:
        try:
            tag, value = info_entry.split('=')
        except:
            #sys.stderr.write("Error in splitting %s by =.\n"%info)
            # ###Error in splitting DS by =.
            continue
        info_tag2value[tag] = value
    
    locus_id = (chromosome, pos)
    refBase = row[col_name2index['REF']]
    altBase = row[col_name2index['ALT']]
    
    altBaseLs = altBase.split(',')
    #altBase could be just "C" or "C,G" (multi-nucleotide)
    alleleLs = [refBase] + altBaseLs
    alleleNumber2Base = {'.':'NA'}
    for i in range(len(alleleLs)):
        alleleNumber2Base[repr(i)] = alleleLs[i]
    
    format_column = row[col_name2index['FORMAT']]
    format_column_ls = format_column.split(':')
    format_column_name2index = getColName2IndexFromHeader(format_column_ls)
        
    if dataEntryType==1:
        data_row = ['NA']*(len(col_index_individual_name_ls)+1)	# extra 1 for the ref
        data_row[0] = refBase
    else:
        data_row = [None]*(len(col_index_individual_name_ls)+1)	# extra 1 for the ref
        data_row[0] = {'GT':refBase, 'DP':-1}
    genotypeCall2Count = {}
    for individual_col_index, individual_name in col_index_individual_name_ls:
        individual_name = individual_name
        if individual_name not in sample_id2index:
            sample_id2index[individual_name] = len(sample_id2index)
        
        #coverage = read_group2coverage[individual_name]
        genotype_data = row[individual_col_index]
        genotype_data_ls = genotype_data.split(':')
        genotype_call_index = format_column_name2index.get('GT')
        genotype_quality_index = format_column_name2index.get('GQ')
        if genotype_quality_index is None:
            genotype_quality_index = format_column_name2index.get('DP')
        depth_index = format_column_name2index.get("DP")
        #GL_index = format_column_name2index.get('GL')
        genotypeCallInBase = 'NA'
        if genotype_call_index is not None and len(genotype_data_ls)>0:
            # or (genotype_call_index is not None and len(genotype_data_ls)<=genotype_call_index):
            # 	#<len(format_column_name2index):
            # #this genotype call is probably empty "./." due to no reads
            #genotype_quality = genotype_data_ls[genotype_quality_index]
            if genotype_call_index is not None and len(genotype_data_ls)>genotype_call_index:
                genotype_call = genotype_data_ls[genotype_call_index]
            else:
                genotype_call = './.'	#missing
            callData = {}
            if genotype_call!='./.' and genotype_call!='.' and genotype_call!='.|.':
                #missing data
                patternSearchResult = diploidGenotypePattern.search(genotype_call)
                if patternSearchResult:
                    allele1 = alleleNumber2Base[patternSearchResult.group(1)]
                    allele2 = alleleNumber2Base[patternSearchResult.group(2)]
                    if allele1!='N' and allele2!='N':
                        genotypeCallInBase = '%s%s'%(allele1, allele2)
                if depth_index is not None:
                    if len(genotype_data_ls)>depth_index:
                        depth = genotype_data_ls[depth_index]
                    else:
                        depth = '.'	#missing DP
                    if depth=='.':	#this means depth=0
                        depth = 0
                    else:
                        depth = int(depth)
                    if minDepth>0 and depth<minDepth:
                        #no read. samtools would still assign ref/ref to this individual
                        genotypeCallInBase = 'NA'	#set it to missing
                    #if depth>maxNoOfReads*coverage or depth<minNoOfReads*coverage:
                    # #2011-3-29 skip. coverage too high or too low
                    #	continue
                    callData['DP'] = depth

        """
        if genotype_call=='0/1' or genotype_call =='1/0':
            #heterozygous, the latter notation is never used though.
            allele = '%s%s'%(refBase, altBase)
            GL_list = genotype_data_ls[GL_index]
            GL_list = GL_list.split(',')
            GL_list = map(float, GL_list)
            GL = GL_list[1]
            sndHighestGL = max([GL_list[0], GL_list[2]])
            deltaGL = GL-sndHighestGL
            
            AD = genotype_data_ls[format_column_name2index.get('AD')]
            AD = map(int, AD.split(','))
            minorAlleleCoverage = min(AD)
            majorAlleleCoverage = max(AD)
            
            if minorAlleleCoverage<=minorAlleleDepthUpperBoundCoeff*coverage and \
                    minorAlleleCoverage>=minorAlleleDepthLowerBoundCoeff*coverage and \
                    majorAlleleCoverage<=majorAlleleDepthUpperBoundCoeff*coverage:
                DP4_ratio = float(AD[0])/AD[1]
                allele = '%s%s'%(refBase, altBase)

        elif genotype_call=='./.' or genotype_call=='.|.':	#missing
            allele = 'NA'
        elif genotype_call =='1/1' or genotype_call =='1|1':
            allele = '%s%s'%(altBase, altBase)
        elif genotype_call =='0/0' or genotype_call=='0|0':
            allele = '%s%s'%(refBase, refBase)
        """
        col_index = sample_id2index.get(individual_name)
        if dataEntryType==1:
            data_row[col_index] = genotypeCallInBase
        else:
            callData['GT'] = genotypeCallInBase
            data_row[col_index] = callData
        if genotypeCallInBase!='NA':
            if genotypeCallInBase not in genotypeCall2Count:
                genotypeCall2Count[genotypeCallInBase] = 0
            genotypeCall2Count[genotypeCallInBase] += 1
    return PassingData(chr=chromosome, chromosome=chromosome, pos=pos,
        position=pos, locus_id=locus_id, quality=quality,
        info_tag2value=info_tag2value,
        refBase=refBase, altBase=altBase,
        alleleLs=alleleLs, alleleNumber2Base=alleleNumber2Base,
        genotypeCall2Count=genotypeCall2Count, data_row=data_row,
        info=info, format=format, filter=filter, vcf_locus_id=vcf_locus_id,
        format_column_name2index=format_column_name2index,
        format_column_ls=format_column_ls)

class VCFRecord(object):
    """
    2011-10-20
        
    """
    def __init__(self, row=None, col_name2index=None,
        individual_name2col_index=None, sample_id2index=None,
        col_index_individual_name_ls=None, minDepth=1,
        sampleStartingColumn=9, ploidy=2, **keywords):
        """
        add argument minDepth
        """
        self.dataEntryType = 2
        # controls whether each genotype data in self.data_row is just genotype (string) or dictionary
        
        self.row = row
        self.col_name2index = col_name2index
        self.individual_name2col_index = individual_name2col_index
        self.sample_id2index = sample_id2index
        self.col_index_individual_name_ls = col_index_individual_name_ls
        #list of [column-index, sample-id], column-index >=9
        self.minDepth = minDepth
        self.sampleStartingColumn = sampleStartingColumn
        self.ploidy = ploidy
        
        
        self.data_row = []
        self.alleleLs = []	#index 0 is refBase, 1 is first altBase, 2 is 2nd altBase .. 
        self.alleleNumber2Base = {}	#map string type of allele 0,1,2 to the actual base
        self.format_column_name2index = None
        self.format_column_ls = None
        
        self._parse(row)
        #
        
    def _parse(self, row):
        """
        2012.5.2
            add refBase, altBase to self
        """
        returnData = parseOneVCFRow(row, self.col_name2index,
            self.col_index_individual_name_ls, self.sample_id2index,
            minDepth=self.minDepth, dataEntryType=self.dataEntryType)
        
        self.chr = returnData.chromosome
        self.chromosome = returnData.chromosome
        self.pos = returnData.pos	#integer
        self.position = returnData.pos	#INTEGER
        self.locus_id = returnData.locus_id
        self.vcf_locus_id = returnData.vcf_locus_id
        self.info_tag2value = returnData.info_tag2value
        self.data_row = returnData.data_row
        #a list of dictionary {'GT': base-call, 'DP': depth}, including one extra sample, the ref
        self.alleleNumber2Base = returnData.alleleNumber2Base
        self.alleleLs = returnData.alleleLs
        
        self.refBase = returnData.refBase
        self.altBase = returnData.altBase
        self.quality = returnData.quality
        self.filter = returnData.filter
        self.info = returnData.info
        self.format = returnData.format
        self.format_column_name2index = returnData.format_column_name2index
        self.format_column_ls = returnData.format_column_ls
    
    def getAAF(self, useMax=True, defaultValue=0):
        """
        argument useMax, which determines to return max or min when there are >1 alternative alleles.
        """
        frequencyLs = self.info_tag2value.get("AF", self.info_tag2value.get("AF1", None))
        if frequencyLs is not None:
            #if more than one alt alleles, AF is a string separated by ",".
            frequencyLs = frequencyLs.split(',')
            frequencyLs = map(float ,frequencyLs)
            if useMax:
                frequency = max(frequencyLs)
            else:
                frequency = min(frequencyLs)
        else:
            frequency = defaultValue
        return frequency
    
    def getGenotypeCallForOneSample(self, sampleID=None):
        """
        2013.05.06
        """
        col_index = self.sample_id2index.get(sampleID)
        if not self.data_row:
            self._parse(self.row)
        return self.data_row[col_index]
    
    def setColumnData(self, columnHeader=None, columnValue=None):
        """
        2013.07.11
            set any column
            columnHeader for chromosome is "CHROM". the initial '#' is removed.
        """
        columnIndex = self.col_name2index.get(columnHeader)
        self.row[columnIndex] = columnValue
    
    def setChromosome(self, chromosome=None):
        """
        2013.07.11
        """
        self.setColumnData(columnHeader="CHROM", columnValue=chromosome)
        self.chromosome = chromosome
        self.chr = chromosome
    
    def setPosition(self, position=None):
        self.setColumnData(columnHeader="POS", columnValue=position)
        self.pos = position
        self.position = position
    
    def setRefAllele(self, allele=None):
        self.setColumnData(columnHeader="REF", columnValue=allele)
        self.refBase = allele
    
    def setAltAllele(self, allele=None):
        self.setColumnData(columnHeader="ALT", columnValue=allele)
        self.altBase = allele
    
    log10Likeli2PhredScaleFunction = lambda x: int(x*-10)
    def setGenotypeCallForOneSample(self, sampleID=None, genotype=None, convertGLToPL=True):
        """
        2913.06.05
        argument genotype should be in VCF format. 'A/G' for unphased;  'A|G' for phased; './.' or '.|.' for missing
        argument convertGLToPL is so far only  for platypus.
            ##FORMAT=<ID=PL,Number=G,Type=Integer,Description="Normalized, 
            #   Phred-scaled likelihoods for genotypes as defined in the VCF specification">
            ##FORMAT=<ID=GL,Number=.,Type=Float,Description="Genotype log10-likelihoods
            #    for AA,AB and BB genotypes, where A = ref and B = variant. Only applicable for bi-allelic sites">
            
            PL = -10*GL
            
            self.log10Likeli2PhredScaleFunction is lambda function for conversion
            convertGLToPL=True by default because TrioCaller or most programs expect a PL, rather than GL.
        """
        #col_index for "ref" is 0. col_index starts from 1 for actual samples.
        col_index = self.sample_id2index.get(sampleID)
        if self.dataEntryType==2:
            self.data_row[col_index]['GT'] = genotype
        else:
            self.data_row[col_index] = genotype
        
        #upon writing out a VCF record, self.row is the one to be written out.
        # self.data_row is for data accessing
        sampleGenotypeAndOtherDataIndex = self.sampleStartingColumn + col_index - 1
        genotypeFieldData = self.row[sampleGenotypeAndOtherDataIndex]
        if genotypeFieldData =='.':
            genotypeFieldDataList = ':'.join(['.']*len(self.format_column_name2index))
        else:
            genotypeFieldDataList = genotypeFieldData.split(':')
        GT_index = self.format_column_name2index.get('GT')
        PL_index = self.format_column_name2index.get('PL')
        ###FORMAT=<ID=PL,Number=G,Type=Integer,Description="Normalized, 
        #   Phred-scaled likelihoods for genotypes as defined in the VCF specification">
        if PL_index is None:
            GL_index = self.format_column_name2index.get('GL')	#from Platypus
            if GL_index is not None:
                ##FORMAT=<ID=GL,Number=.,Type=Float,Description="Genotype log10-likelihoods
                #   for AA,AB and BB genotypes, where A = ref and B = variant. Only applicable for bi-allelic sites">
                if convertGLToPL:	# so far only  for platypus
                    GL = genotypeFieldDataList[GL_index]
                    GL_list = map(float, GL.split(','))
                    PL_list = map(self.log10Likeli2PhredScaleFunction, GL_list)
                    PL = ','.join(PL_list)
                    if self.format.find('GL')>=0:	#change the field name in format as well
                        self.format = self.format.replace('GL', 'PL')
                else:
                    PL = genotypeFieldDataList[GL_index]	#this is kind of wrong
            else:
                PL = '.,.,.'
            PL_index = GL_index
        else:
            #if genotype=='./.' or genotype=='.|.':
            #	PL = '.,.,.'
            #else:
            PL = genotypeFieldDataList[PL_index]
        
        if PL=='.':	#should be .,.,.
            PL = '.,.,.'
        if PL_index is not None:
            genotypeFieldDataList[PL_index] = PL
        genotypeFieldDataList[GT_index] = genotype
        self.row[sampleGenotypeAndOtherDataIndex] = ':'.join(genotypeFieldDataList)
    
    def getGenotypeCallForAllSamples(self):
        """
        2013.05.06
        """
        if not self.data_row:
            self._parse(self.row)
        return self.data_row
    
class VCFFile(object):
    """
    self.reader = VCFFile(inputFname=self.originalVCFFname, mode='r')
    
    
    self.writer = VCFFile(outputFname=self.outputFname, mode='w')
    self.writer.metaInfoLs = self.reader.metaInfoLs
    self.writer.header = self.reader.header
    self.writer.writeMetaAndHeader()
    
    for vcfRecord in self.reader:
        for sampleID, sample_index in vcfRecord.sample_id2index.items():
            beagleFile = self.sampleID2BeagleFile.get(sampleID)
            beagleGenotype = beagleFile.getGenotypeOfOneSampleOneLocus(sampleID=sampleID, locusID=None)
            vcfRecord.setGenotypeCallForOneSample(sampleID=sampleID,
                genotype='%s|%s'%(beagleGenotype[0], beagleGenotype[1]))
            counter += 1
        self.writer.writeVCFRecord(vcfRecord)
    
    
    self.writer = VCFFile(outputFname=self.outputFname, mode='w')
    self.writer.makeupHeaderFromSampleIDList(sampleIDList=snpData.row_id_ls)
    self.writer.writeMetaAndHeader()
    """
    
    __doc__ = __doc__
    option_default_dict = {
        ('inputFname', 0, ): [None, 'i', 1, 'a VCF input file to read in the VCF content.'],\
        ('outputFname', 0, ): [None, 'o', 1, 'a VCF output file to hold the the VCF content.'],\
        ('mode', 1, ): ['rb', '', 1, 'rb: bam file. r: sam file.'],\
        ('minorAlleleDepthLowerBoundCoeff', 1, float): [1/4., 'M', 1, 
            'minimum read depth multiplier for an allele to be called (heterozygous or homozygous)', ],\
        ('minorAlleleDepthUpperBoundCoeff', 1, float): [3/4., 'A', 1,
            'maximum read depth multiplier for the minor allele of a heterozygous call', ],\
        ('majorAlleleDepthUpperBoundCoeff', 1, float): [7/8., 'a', 1,
            'maximum read depth multiplier for the major allele of het call'],\
        ('maxNoOfReadsForGenotypingError', 1, float): [1, 'x', 1, 
            'if read depth for one allele is below or equal to this number, regarded as genotyping error ', ],\
        ('depthUpperBoundCoeff', 1, float): [2, 'm', 1,
            'depthUpperBoundCoeff*coverage = maximum read depth for one base'],\
        ('depthLowerBoundCoeff', 1, float): [1/4., 'O', 1,
            'depthLowerBoundCoeff*coverage = minimum read depth multiplier for one base'],\
        ('depthUpperBoundMultiSampleCoeff', 1, float): [3, 'N', 1, 
            'across n samples, ignore bases where read depth > coverageSum*depthUpperBoundCoeff*multiplier.'],\
        ('seqCoverageFname', 0, ): ['', 'q', 1,
            'The sequence coverage file. tab/comma-delimited: individual_sequence.id coverage'],\
        ('minDepth', 0, float): [0, '', 1, 'minimum depth for a call to regarded as non-missing', ],\
        ('ploidy', 0, int): [2, '', 1, 'ploidy of the organism from which the genotype is from', ],\
        ('defaultCoverage', 1, float): [5, 'd', 1,
            'default coverage when coverage is not available for a read group'],\
        ("site_type", 1, int): [1, 's', 1, '1: all sites, 2: variants only'],\
        ('debug', 0, int):[0, 'b', 0, 'toggle debug mode'],\
        ("sampleStartingColumn", 1, int): [9, '', 1,
            'The column index (starting from 0) of the 1st sample genotype'],\
        ('report', 0, int):[0, 'r', 0, 'toggle report, more verbose stdout/stderr.']
        }
    
    def __init__(self, **keywords):
        """
        2012.9.5 set default minDepth=0
        2011-9-27
        """
        self.ad = ProcessOptions.process_function_arguments(keywords,
            self.option_default_dict, error_doc=self.__doc__, \
            class_to_have_attr=self)
        
        self.header = None
        # the list of column headers (the header line starting by #CHROM)
        self.headerWithoutHash = None
        # same as self.header except, instead of "#CHROM", it is "CHROM".
        self.sample_id_ls = []
        self.sample_id2index = {}
        #the index is the index of its column in the genotype_call_matrix
        self.locus_id_ls = []
        self.locus_id2row_index = {}
        self.locus_id2data = {}
        self.genotype_call_matrix = []
        self.col_name2index = {}	#column index in file
        self.col_index_individual_name_ls = None
        self.individual_name2col_index = {}
        #not the matrix column, the column in input file
        self.metaInfoLs = []
        # anything before the "#CHROM" line. each entry is a raw line content, including '\n'
        self.sampleIDHeader = []
        # a list of sample column headers (from sampleStartingColumn)
        
        self.inf = None
        self.reader = None
        self._initializeInput(self.inputFname)
        
        self.outf = None
        self.writer = None
        self._initializeOutput(self.outputFname)
    
    def _initializeInput(self, inputFname=None):
        """
        """
        if inputFname and self.mode[0]=='r':
            self.inf = utils.openGzipFile(inputFname, mode='r')
            """
            if inputFname[-3:]=='.gz':
                import gzip
                self.inf = gzip.open(inputFname, 'rb')
            else:
                self.inf = open(inputFname)
            """
            self.reader =csv.reader(self.inf, delimiter='\t')
            self._parseHeader()
    
    def _initializeOutput(self, outputFname=None):
        """
        handle gzipped file as well.
        """
        if outputFname:
            if outputFname[-3:]=='.gz':
                import gzip
                self.outf = gzip.open(outputFname, 'wb')
            else:
                self.outf = open(outputFname, 'w')
            self.writer = csv.writer(self.outf, delimiter='\t')
    
    def _getIndividual2ColIndex(self, header, col_name2index=None, sampleStartingColumn=9):
        """
        2012.10.5 add "_" in front of the function name
        called by other function, not meant for public
        """
        if self.report:
            sys.stderr.write("\t Finding all individuals ...")
        no_of_cols = len(header)
        individual_name2col_index = {}
        #individual's column name -> an opened file handler to store genetic data
        col_index_individual_name_ls = []
        
        counter = 0
        for i in range(sampleStartingColumn, no_of_cols):
            individualName = header[i].strip()
            col_index = col_name2index.get(individualName)
            if not individualName:	#ignore empty column
                continue
            if individualName[:-4]=='.bam':
                individualCode = individualName[:-4]
                #get rid of .bam
            else:
                individualCode = individualName
            individual_name2col_index[individualCode] = col_index
            col_index_individual_name_ls.append([col_index, individualCode])
            counter += 1
        self.individual_name2col_index = individual_name2col_index
        
        col_index_individual_name_ls.sort()	#sorted by column index
        if self.report:
            sys.stderr.write("%s individuals added. Done.\n"%(counter))
        return col_index_individual_name_ls
    
    def get_col_index_individual_name_ls(self):
        """
        2012.10.5 accessor of self.col_index_individual_name_ls
            a list of [column-index, individual_name].
                The individual_name excludes the trailing ".bam" if there is one.
            The column-index starts from the sampleStartingColumn (=9 usually). 
        """
        return self.col_index_individual_name_ls
    
    def getIndividualName2ColIndex(self):
        """
        2012.10.5 public version of _getIndividual2ColIndex()
            a dictionary between samplID and column-index (which column in the VCF file)
            
            difference between sampleID and individual_name
                the latter excludes the trailing ".bam" if there is one. 
            
            The column-index starts from the sampleStartingColumn (=9 usually).
        """
        return self.individual_name2col_index
    
    def getSampleID2DataIndex(self):
        """
        2012.10.5 accessor of self.sample_id2index
            It includes "ref" as index=0.
            The index is the index of each sample in the variant data_matrix or data_row of VCFRecord.
            
        """
        
        return self.sample_id2index
    
    def parseFile(self):
        """
        #2012.8.8 bug fix, same locus appearing >1 times
        2011.11.30
            bugfix: genotype_call could be ./.(missing) and depth=. (missing) in "./.:.:.:.:.".
        2011-9-27
            based on discoverFromVCF() from vervet/src/GenotypeCallByCoverage.py
            
            It reads all genotype calls into the matrix so its memory footprint might explode.
        """
        sys.stderr.write("Parsing %s ... \n"%(os.path.basename(self.inputFname),))
        
        locus_id2row_index = self.locus_id2row_index
        
        """
        writer = csv.writer(open(outputFname, 'w'), delimiter='\t')
        header = ['sample', 'snp_id', 'chr', 'pos', 'qual', 'DP', 'minDP4', 'DP4_ratio', 'MQ']
        moreHeader = ['GQ', 'GL', 'SB', 'QD', 'sndHighestGL', 'deltaGL']
        #['AF', 'AC','AN', 'Dels', 'HRun', 'HaplotypeScore','MQ0', 'QD']	#2011-3-4 useless
        if VCFOutputType==2:
            header += moreHeader
        chr_pure_number_pattern = re.compile(r'[a-z_A-Z]+(\d+)')
        chr_number_pattern = re.compile(r'chr(\d+)')
        """
        
        individual_name2col_index = self.individual_name2col_index
        col_name2index = self.col_name2index
        counter = 0
        real_counter = 0
        
        for row in self.reader:
            counter += 1
            returnData = parseOneVCFRow(row, self.col_name2index,
                self.col_index_individual_name_ls, self.sample_id2index, \
                minDepth=self.minDepth, dataEntryType=1)
            
            genotypeCall2Count = returnData.genotypeCall2Count
            data_row = returnData.data_row
            
            current_locus = returnData.locus_id
            if current_locus not in self.locus_id2data:
                self.locus_id2data[current_locus] = returnData.info_tag2value
                self.locus_id2data[current_locus]['REF'] = returnData.refBase
                self.locus_id2data[current_locus]['ALT'] = returnData.altBase
            
            if len(genotypeCall2Count)>self.site_type-1:
                #whether polymorphic across samples or all sites in vcf
                real_counter += 1
                if current_locus in locus_id2row_index:
                    sys.stderr.write("WARNING: Locus %s at line %s already in locus_id2row_index with row_index = %s. Skip\n"%\
                        (repr(current_locus), counter, locus_id2row_index.get(current_locus)))
                else:
                    locus_id2row_index[current_locus] = len(locus_id2row_index)
                    self.locus_id_ls.append(current_locus)
                    self.genotype_call_matrix.append(data_row)
                    real_counter += 1
            
            if counter%2000==0 and self.report:
                sys.stderr.write("%s\t%s\t%s"%("\x08"*80, counter, real_counter))
        sys.stderr.write("%s loci X %s samples.\n"%(len(self.locus_id_ls), len(self.sample_id_ls)))
    
    def _parseHeader(self):
        """
        add all header content into self.metaInfoLs
            except the last header line, which goes into self.sampleIDHeader
        this function is run inside __init__().
        """
        self.metaInfoLs = []
        # anything before the "#CHROM" line. each entry is a raw line content, including '\n'
        self.sampleIDHeader = []
        # a list of column headers (#CHROM)
        
        self.sample_id2index['ref'] = 0
        #ref is at column 0. "ref" must not be equal to any read_group.
        self.sample_id_ls.append('ref')
        
        """
        writer = csv.writer(open(outputFname, 'w'), delimiter='\t')
        header = ['sample', 'snp_id', 'chr', 'pos', 'qual', 'DP', 'minDP4', 'DP4_ratio', 'MQ']
        moreHeader = ['GQ', 'GL', 'SB', 'QD', 'sndHighestGL', 'deltaGL']
        #['AF', 'AC','AN', 'Dels', 'HRun', 'HaplotypeScore','MQ0', 'QD']	#2011-3-4 useless
        if VCFOutputType==2:
            header += moreHeader
        chr_pure_number_pattern = re.compile(r'[a-z_A-Z]+(\d+)')
        chr_number_pattern = re.compile(r'chr(\d+)')
        """
        
        counter = 0
        real_counter = 0
        
        for line in self.inf:
            if line[:6]=='#CHROM':
                line = line.strip()
                #get rid of the trailing \n
                row = line.split('\t')
                self.sampleIDHeader = row[self.sampleStartingColumn:]
                self.header = row[:]
                self.headerWithoutHash= row[:]
                self.headerWithoutHash[0] = 'CHROM'	#discard the #
                self.col_name2index = getColName2IndexFromHeader(
                    self.headerWithoutHash, skipEmptyColumn=True)
                self.col_index_individual_name_ls = self._getIndividual2ColIndex(
                    self.headerWithoutHash, self.col_name2index)
                for individual_col_index, individual_name in self.col_index_individual_name_ls:
                    read_group = individual_name.strip()
                    if read_group not in self.sample_id2index:
                        self.sample_id2index[read_group] = len(self.sample_id2index)
                        self.sample_id_ls.append(read_group)
                break	# "#CHROM" is the last line of the self.headerWithoutHash
            elif line[0]=='#':
                self.metaInfoLs.append(line)
                #continue
            else:	#leave everything for parseFile or parseIter
                break
    
    def constructColName2IndexFromHeader(self):
        """
        2013.09.05
            to be compatible with MatrixFile
        """
        return self.col_name2index
    
    def getHeader(self):
        """
        2013.09.05
            to be compatible with MatrixFile
        """
        return self.header
    
    def parseIter(self):
        """
        2011-11-2
            an iterator over each line (call data) in VCF
        """
        for row in self.reader:
            vcfRecord = VCFRecord(row, col_name2index=self.col_name2index, \
                individual_name2col_index=self.individual_name2col_index,
                sample_id2index=self.sample_id2index,\
                col_index_individual_name_ls=self.col_index_individual_name_ls,\
                minDepth=self.minDepth,
                sampleStartingColumn=self.sampleStartingColumn,
                ploidy=self.ploidy)
            yield vcfRecord
    
    def __iter__(self):
        """
        2012.8.25 make itself an iterator
        """
        return self
    
    def next(self):
        """
        2012.8.25 make itself an iterator
        """
        try:
            row = next(self.reader)
            if row is None or row==[] or row=="":
                raise StopIteration
            vcfRecord = VCFRecord(row, col_name2index=self.col_name2index,
                individual_name2col_index=self.individual_name2col_index,
                sample_id2index=self.sample_id2index,
                col_index_individual_name_ls=self.col_index_individual_name_ls,
                minDepth=self.minDepth,
                sampleStartingColumn=self.sampleStartingColumn,
                ploidy=self.ploidy)
            return vcfRecord
        except StopIteration:
            raise StopIteration
        except:
            sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
            import traceback
            traceback.print_exc()
            raise
    
    def _countHomoHetCallsForEachSampleFromVCF(self):
        """
        2011-11-2
            given a VCF file, count the number of homo-ref, homo-alt, het calls
            
        """
        sys.stderr.write("Count the number of homozygous-ref/alt & het.\n")
        sampleID2data = {}
        #key is sampleID, value is a list of 3 numbers. 'NoOfHomoRef', 'NoOfHomoAlt', 'NoOfHet'
        
        no_of_total = 0.
        minStart = None
        for vcfRecord in self.parseIter():
            locus_id = vcfRecord.locus_id
            chromosome = vcfRecord.chromosome
            pos = vcfRecord.pos
            pos = int(pos)
            refBase = vcfRecord.data_row[0].get("GT")[0]
            
            for sample_id, sample_index in self.sample_id2index.items():
                if sample_id=='ref':	#ignore the reference
                    continue
                if sample_id not in sampleID2data:
                    sampleID2data[sample_id] = [0, 0, 0]
                if not vcfRecord.data_row[sample_index]:	#None for this sample
                    continue
                callForThisSample = vcfRecord.data_row[sample_index].get('GT')
                if not callForThisSample or callForThisSample=='NA':
                    continue
                if callForThisSample[0]==refBase and callForThisSample[1]==refBase:
                    #homozygous reference allele
                    sampleID2data[sample_id][0]+=1
                elif callForThisSample[0]==callForThisSample[1] and callForThisSample[0]!=refBase:
                    #homozygous alternative allele
                    sampleID2data[sample_id][1]+=1
                elif callForThisSample[0]!=callForThisSample[1]:
                    sampleID2data[sample_id][2]+=1
        
        sampleIDLs = sorted(sampleID2data)
        for sampleID in sampleIDLs:
            count_data = sampleID2data.get(sampleID)
            noOfHomoRef, noOfHomoAlt, noOfHet = count_data[:3]
            no_of_calls = float(sum(count_data))
            if no_of_calls>0:
                fractionOfHomoAlt = noOfHomoAlt/no_of_calls
                fractionOfHet = noOfHet/no_of_calls
            else:
                fractionOfHomoAlt = -1
                fractionOfHet = -1
        sys.stderr.write("Done.\n")
    
    def getSampleIDList(self):
        """
        2013.07.17 get data from self.header, not from self.sample_id_ls anymore
        2012.7.18
            self.sample_id_ls includes a "ref" in index=0 cell.
                this function helps to reduce confusion by returning
            a list of sample IDs without "ref"
        """
        return self.header[self.sampleStartingColumn:]
    
    def getNoOfLoci(self):
        """
        2012.9.25
            fast way to get the number of loci, no VCF parsing.
        """
        if self.inf is None:
            return None
        
        if len(self.sample_id_ls)==0 and len(self.sampleIDHeader)==0:
            self._parseHeader()
        no_of_loci = 0
        for line in self.inf:
            no_of_loci += 1
        self._resetInput()
        
        return no_of_loci
    
    def _resetInput(self):
        """
        run only to reset this VCFFile object and the read cursor is at the beginning of all VCF records.
        split out of getNoOfLoci()
        """
        #reset the inf
        self.inf.seek(0)
        #run this to move the cursor to where the VCF records start
        self._parseHeader()
    
    def _makeupMetaInfo(self):
        """
        2013.07.03
            for writing VCF files
        """
        if self.metaInfoLs is None or self.metaInfoLs==[]:
            self.metaInfoLs = []
            self.metaInfoLs.append("##fileformat=VCFv4.0\n")
    
    def makeupHeaderFromSampleIDList(self, sampleIDList=None):
        """
        2013.07.03
            for writing VCF files
        #CHROM  POS     ID      REF     ALT     QUAL    FILTER  INFO    FORMAT  
        """
        self.header = ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT"] + sampleIDList
        return self.header
        
    def writeMetaAndHeader(self, outf=None):
        """
        2012.5.10
            vcfFile = VCFFile(outputFname=...)
            vcfFile.metaInfoLs = oldVCFFile.metaInfoLs
            vcfFile.header = [...]
            vcfFile.writeMetaAndHeader()
            
            write out the self.header and self.metaInfoLs
        """
        if outf is None:
            outf = self.outf
            writer = self.writer
        else:
            writer = csv.writer(outf, delimiter='\t')
        if outf is None:
            sys.stderr.write("Warning: outf is None. nothing is written out.\n")
        else:
            if not self.metaInfoLs:
                self._makeupMetaInfo()
            for metaInfo in self.metaInfoLs:
                outf.write(metaInfo)
            writer.writerow(self.header)
    
    def writeVCFRecord(self, vcfRecord=None, outf=None):
        """
        2012.5.10
        """
        if outf is None:
            outf = self.outf
            writer = self.writer
        else:
            writer = csv.writer(outf, delimiter='\t')
        if outf is None:
            sys.stderr.write("Warning: outf is None. nothing is written out.\n")
        else:
            self.locus_id_ls.append(vcfRecord.vcf_locus_id)
            data_row = [vcfRecord.chromosome, vcfRecord.pos, vcfRecord.vcf_locus_id, \
                vcfRecord.refBase, vcfRecord.altBase, \
                vcfRecord.quality, vcfRecord.filter,\
                vcfRecord.info, vcfRecord.format] + vcfRecord.row[self.sampleStartingColumn:]
            writer.writerow(data_row)
    
    def writerow(self, data_row):
        """
        """
        self.writer.writerow(data_row)
    
    def close(self,):
        """
        """
        if self.inf:
            del self.inf, self.reader
        if self.outf:
            del self.outf, self.writer
    
    def getLocus2AlternativeAlleleFrequency(self):
        """
        adapted from VariantDiscovery.getContig2Locus2Frequency()
        """
        locus2frequency = {}
        for vcfRecord in self:
            current_locus = (vcfRecord.chromosome, vcfRecord.pos)
            AAF = vcfRecord.getAAF()
            locus2frequency[current_locus] = AAF
            counter += 1
        self._resetInput()
        return locus2frequency


#!/usr/bin/env python3
"""

Examples:
	#postgresql
	TAIRGeneXML2GenomeDB.py -c -i *.xgs
	
	TAIRGeneXML2GenomeDB.py -z localhost -k genome -c -i TAIR9_genome_release/Tair9_XML/ch1.xml -r

	#mysql
	TAIRGeneXML2GenomeDB.py -v mysql -u yh -z papaya -d genome_tair -k "" -c -i TAIR9_genome_release/Tair9_XML/ch1.xml -r
	
	TAIRGeneXML2GenomeDB.py -v mysql -u yh -z banyan -d genome_tair -r  ~/script/variation/data/TAIR9/TAIR9_genome_release/Tair9_XML/ch[2-5].*
	
Description:
	Put TAIR xml genome release files into db.
	This program checks whether an entry is in db or not only for AnnotAssembly, Gene, EntrezgeneMapping, EntrezgeneType,
		Gene2go, GeneCommentaryType, but not for GeneCommentary and GeneSegment.
	If program stops half way and needs to be re-run, all prior inserted data has to be purged MANUALLY.
	
	xml files are downloaded from ftp://ftp.arabidopsis.org/home/tair/Genes/TAIR9_genome_release/Tair9_XML/
"""

import sys, getopt, os
import xml.etree.cElementTree as ElementTree
from palos.db.GenomeDB import GenomeDatabase, Gene, SequenceType, EntrezgeneType, \
	GeneSegment, GeneCommentaryType, GeneCommentary, AnnotAssembly, Gene2go
from palos import PassingData
import datetime
#import argparse	#in python 2.7
from optparse import OptionParser

def usernameOptionCallBack(option, opt_str, value, parser):
	#2010-8-13
	# useless. called only when -u is supplied, which misses the whole point.
	option_name = option.dest
	if not value:
		value = input("%s: "%opt_str)
	setattr(parser.values, option_name, value)

def passwordOptionCallBack(option, opt_str, value, parser):
	option_name = option.dest
	import getpass
	if not value:
		value = getpass.getpass("%s: "%opt_str)
	setattr(parser.values, option_name, value)


class TAIRGeneXML2GenomeDB:
	__doc__ = __doc__
	
	usage = "usage: %prog [options] inputfiles \n\n-h/--help for help."
	parser = OptionParser(usage=usage, description='', epilog=__doc__)
	#parser.add_option('inputfiles', type=int, nargs='+',
	#				help='input files in XML. can detect gzipped file by .gz in filename.')
	parser.add_option('-v', '--drivername', default="postgres",
					help='which type of database? mysql or postgres [%default]')
	parser.add_option('-z', '--hostname', dest='hostname', default="localhost",
					help='hostname of the db server [%default]')
	parser.add_option('-d', '--dbname', dest='dbname', default="graphdb",
					help='database name [%default]')
	parser.add_option('-k', '--schema', dest='schema', default="",
					help='database schema (only for postgres) [%default]')
	parser.add_option('-u', '--db_user', dest='db_user', 
					help='database username [%default]')
	parser.add_option('-p', '--db_passwd', 
					help='database password [%default]')
	parser.add_option('-x', '--tax_id', default=3702, type=int,
					help='taxonomy ID for the organism [%default]')
	parser.add_option('-c', '--commit', default=False, action='store_true',
					help='commit db transaction. The program commits regardless of this argument.[%default]')
	parser.add_option('-b', '--debug', default=False, action='store_true',
					help='toggle debug mode [%default]')
	parser.add_option('-r', '--report', default=False, action='store_true',
					help='toggle report, more verbose stdout/stderr. [%default]')
	
	option_default_dict = {('drivername', 1,):['postgres', 'v', 1, 'which type of database? mysql or postgres', ],\
							('hostname', 1, ): ['localhost', 'z', 1, 'hostname of the db server', ],\
							('dbname', 1, ): ['graphdb', 'd', 1, 'database name', ],\
							('schema', 0, ): ['', 'k', 1, 'database schema name', ],\
							('db_user', 1, ): [None, 'u', 1, 'database username', ],\
							('db_passwd', 1, ): [None, 'p', 1, 'database password', ],\
							('tax_id', 1, int): [3702, '', 1, 'taxonomy ID', ],\
							('inputfiles', 1, ):[None, 'i', 1, 'comma-separated input filenames, or a list of files'],\
							('commit', 0, int):[0, 'c', 0, 'commit db transaction'],\
							('debug', 0, int):[0, 'b', 0, 'toggle debug mode'],\
							('report', 0, int):[0, 'r', 0, 'toggle report, more verbose stdout/stderr.']}
	def __init__(self,  argv):
		"""
		2008-07-29
			use option_default_dict
		"""
		(options, args) = self.parser.parse_args(argv)
		for option in self.parser.option_list:
			option_name = option.dest
			if option_name:
				setattr(self, option_name, getattr(options, option_name))
		self.inputfiles = args[1:]
		if len(self.inputfiles) <1:
			self.parser.error("Please supply input files.\n")
		
		#from palos import ProcessOptions
		#self.ad = ProcessOptions.process_function_arguments(keywords, self.option_default_dict, error_doc=self.__doc__, class_to_have_attr=self)
		
		#if type(self.inputfiles)==str:
		#	self.inputfiles = self.inputfiles.split(',')
	
	def getGeneCommentaryType(self, session, type_name=None, type_id=None):
		"""
		2010-8-12
		"""
		if type_id is not None:
			gene_commentary_type = GeneCommentaryType.get(type_id)
		else:
			gene_commentary_type = GeneCommentaryType.query.filter_by(type=type_name).first()
		if not gene_commentary_type:
			if self.debug:
				sys.stderr.write("\t Gene-commentary_type=%s not in db yet.\n"%type_name)
			gene_commentary_type = GeneCommentaryType(type=type_name)
			if type_id is not None:
				gene_commentary_type.id = type_id
			session.add(gene_commentary_type)
			session.flush()
		return gene_commentary_type
	
	def addGeneCommentary(self, element, gene, session, param_obj=None):
		"""
		2010-8-12
		"""
		if not hasattr(param_obj, 'no_of_total'):
			setattr(param_obj, 'no_of_total', 0)
		if not hasattr(param_obj, 'no_of_into_db'):
			setattr(param_obj, 'no_of_into_db', 0)
		
		model_start = int(element.findtext('COORDSET/END5'))	#2010-8-18 this usually corresponds to the CDS start-stop, not EXONs
		model_stop = int(element.findtext('COORDSET/END3'))
		pub_locus = element.findtext('PUB_LOCUS')
		
		
		CDS_box_ls = []
		EXON_box_ls = []
		UTR_box_ls = []
		for exon_element in element.findall('EXON'):
			#gene segment for this exon
			start = int(exon_element.findtext('COORDSET/END5'))
			stop = int(exon_element.findtext('COORDSET/END3'))
			EXON = (min(start, stop), max(start, stop))
			EXON_box_ls.append(EXON)
			
			# if CDS exist, get all boxes, construct another gene commentary, and each CDS is a gene segment
			for cds_element in exon_element.findall("CDS"):
				start = int(cds_element.findtext('COORDSET/END5'))
				stop = int(cds_element.findtext('COORDSET/END3'))
				CDS = (min(start, stop), max(start, stop))	#always smaller as start. strand information in EntrezgeneMapping
				CDS_box_ls.append(CDS)
			utr_elements = exon_element.find("UTRS")
			if utr_elements:
				for utr_element in utr_elements:
					start = int(utr_element.findtext('COORDSET/END5'))
					stop = int(utr_element.findtext('COORDSET/END3'))
					UTR = (min(start, stop), max(start, stop))	#always smaller as start. strand information in EntrezgeneMapping
					UTR_box_ls.append(UTR)
		
		EXON_box_ls.sort()
		if len(EXON_box_ls)>0:	#2010-8-18 replace the CDS start-stop with EXON start-stop.
			model_start = EXON_box_ls[0][0]
			model_stop = EXON_box_ls[-1][1]
		gene_commentary = GeneCommentary(label=pub_locus, \
						text=gene.description, start=model_start, stop=model_stop)
		# start is always smaller than stop. Strand information is in table EntrezgeneMapping
		gene_commentary.gene = gene
		
		commentary_type_name = gene.entrezgene_type
		
		gene_commentary_type = self.getGeneCommentaryType(session, type_name=commentary_type_name)
		
		gene_commentary.gene_commentary_type = gene_commentary_type
		session.add(gene_commentary)
		
		# 2010-8-18 get all the introns
		no_of_exons = len(EXON_box_ls)
		for i in range(no_of_exons):
			EXON = EXON_box_ls[i]
			start, stop = EXON[:2]
			gene_segment = GeneSegment(start=min(start, stop), stop=max(start, stop))
			gene_segment.gene_commentary = gene_commentary
			gene_commentary_type = self.getGeneCommentaryType(session, type_name='exon')
			gene_segment.gene_commentary_type = gene_commentary_type
			session.add(gene_segment)
			param_obj.no_of_into_db += 1
			
			if i>0:
				previous_EXON = EXON_box_ls[i-1]
				current_EXON = EXON_box_ls[i]
				intron_start = previous_EXON[1]+1
				intron_stop = current_EXON[0]-1
				
				gene_segment = GeneSegment(start=intron_start, stop=intron_stop)
				gene_segment.gene_commentary = gene_commentary
				gene_commentary_type = self.getGeneCommentaryType(session, type_name='intron')
				gene_segment.gene_commentary_type = gene_commentary_type
				session.add(gene_segment)
				param_obj.no_of_into_db += 1
		
		
		CDS_box_ls.sort()
		if CDS_box_ls:
			whole_CDS_start = CDS_box_ls[0][0]
			whole_CDS_stop = CDS_box_ls[-1][1]
			protein_commentary = GeneCommentary(label=pub_locus, \
						start=whole_CDS_start, stop=whole_CDS_stop)
			# start is always smaller than stop. Strand information is in table EntrezgeneMapping
			protein_commentary.gene = gene
			protein_commentary.gene_commentary = gene_commentary
			
			gene_commentary_type = self.getGeneCommentaryType(session, type_name="peptide")
			
			protein_commentary.gene_commentary_type = gene_commentary_type
			session.add(protein_commentary)
			param_obj.no_of_into_db += 1
			
			for CDS in CDS_box_ls:
				gene_segment = GeneSegment(start=CDS[0], stop=CDS[1])
				gene_segment.gene_commentary = protein_commentary
				gene_commentary_type = self.getGeneCommentaryType(session, type_name='CDS')
				gene_segment.gene_commentary_type = gene_commentary_type
				session.add(gene_segment)
				param_obj.no_of_into_db += 1
			
			if UTR_box_ls:
				#2010-8-18 add all UTR boxes to this protein_commentary
				for UTR in UTR_box_ls:
					if gene.strand =='+1':
						if UTR[1]<whole_CDS_start:
							type_name = '5UTR'
						elif UTR[0]>whole_CDS_stop:
							type_name = '3UTR'
						else:
							pass	#shouldn't be here
					else:
						if UTR[1]<whole_CDS_start:
							type_name = '3UTR'
						elif UTR[0]>whole_CDS_stop:
							type_name = '5UTR'
						else:
							pass	#shouldn't be here
					gene_segment = GeneSegment(start=UTR[0], stop=UTR[1])
					gene_segment.gene_commentary = protein_commentary
					gene_commentary_type = self.getGeneCommentaryType(session, type_name=type_name)
					gene_segment.gene_commentary_type = gene_commentary_type
					session.add(gene_segment)
					param_obj.no_of_into_db += 1
	
	def addGene(self, session, annot_assembly, element, type_of_gene=None, param_obj=None):
		"""
		2011-6-25
			upgraded to the new GenomeDB schema, but not tested
		2010-8-11
		"""
		if not hasattr(param_obj, 'no_of_total'):
			setattr(param_obj, 'no_of_total', 0)
		if not hasattr(param_obj, 'no_of_into_db'):
			setattr(param_obj, 'no_of_into_db', 0)
		
		gene_synonym = element.find("GENE_SYNONYM")
		feat_name = element.findtext("FEAT_NAME")
		com_name = element.findtext("COM_NAME")
		start = int(element.findtext('COORDSET/END5'))
		stop = int(element.findtext('COORDSET/END3'))
		
		date_string = element.findtext("DATE")
		try:	#2010-8-12
			modification_date = datetime.datetime.strptime(date_string, '%b %d %Y %I:%M%p')
		except:
			modification_date = None
		
		if start<=stop:
			strand = "+1"
		else:
			# reverse start and stop
			strand = "-1"
			tmp = start
			start = stop
			stop = tmp
		
		gene_info = element.find("GENE_INFO")
		
		locustag = gene_info.findtext("PUB_LOCUS")
		gene_symbol = gene_info.findtext("GENE_SYM")
		if not gene_symbol:
			gene_symbol = locustag
		com_name = gene_info.findtext("COM_NAME")
		comment = gene_info.findtext("COMMENT")
		pub_comment = gene_info.findtext("PUB_COMMENT")
		if comment:
			pub_comment += "\n %s"%comment
		if type_of_gene=='protein-coding':
			is_pseudogene = gene_info.findtext("IS_PSEUDOGENE")
			if is_pseudogene=='1':
				type_of_gene = 'pseudo'
		else:
			type_of_gene = element.tag	#non-protein-coding genes have their tags as gene types (like rRNA,)
		
		entrezgene_type = EntrezgeneType.query.filter_by(type=type_of_gene).first()	#query the db to see if it exists or not
		if not entrezgene_type:
			entrezgene_type = EntrezgeneType(type=type_of_gene)
			session.add(entrezgene_type)
			session.flush()
			param_obj.no_of_into_db += 1
		
		gene = Gene.query.filter_by(tax_id=annot_assembly.tax_id).filter_by(chromosome=annot_assembly.chromosome).\
			filter_by(locustag=locustag).filter_by(strand=strand).filter_by(start=start).filter_by(stop=stop).\
			filter_by(type_of_gene=type_of_gene).first()
		if not gene:
			gene = Gene(gene_symbol=gene_symbol, tax_id=annot_assembly.tax_id, locustag=locustag, \
				chromosome=annot_assembly.chromosome,\
				description = pub_comment, full_name_from_nomenclature_authority=com_name, type_of_gene=type_of_gene, \
				modification_date=modification_date, strand=strand, start=start, stop=stop,)
			gene.annot_assembly = annot_assembly
			gene.entrezgene_type = entrezgene_type
			session.add(gene)
		else:
			param_obj.no_of_genes_already_in_db += 1
		
		gene_ontology  = gene_info.find("GENE_ONTOLOGY")
		if gene_ontology:
			for go_id_elem in gene_ontology.findall('GO_ID'):
				go_id = go_id_elem.get("ASSIGNMENT")
				evidence = go_id_elem.find("GO_EVIDENCE/EV_CODE").get("CODE")
				evidence_source = go_id_elem.findtext("GO_EVIDENCE/EVIDENCE")
				go_term = go_id_elem.findtext("GO_TERM")
				category = go_id_elem.findtext("GO_TYPE").split()[1]	#component in "cellular component" is enough
				if gene.gene_id:
					gene2go = Gene2go.query.filter_by(gene_id=gene.gene_id).filter_by(go_id=go_id).filter_by(evidence=evidence).\
						filter_by(category=category).filter_by(tax_id=annot_assembly.tax_id).first()
				else:
					gene2go = None
				if not gene2go:
					gene2go = Gene2go(tax_id=annot_assembly.tax_id, go_id=go_id, evidence=evidence, category=category, \
									go_description=go_term, pubmed_ids=evidence_source)
					gene2go.gene = gene
					session.add(gene2go)
					session.flush()
					param_obj.no_of_into_db += 1
				else:
					gene2go.pubmed_ids += '; %s'%evidence_source
					param_obj.no_of_gene2go_already_in_db += 1
		"""
		if gene.gene_id:
			entrezgene_mapping = Gene.query.filter_by(gene_id=gene.gene_id).first()
		else:
			entrezgene_mapping = None
		if not entrezgene_mapping:
			entrezgene_mapping = EntrezgeneMapping(tax_id=annot_assembly.tax_id, chromosome=annot_assembly.chromosome,\
											start=start, stop=stop, strand=strand,)
			entrezgene_mapping.annot_assembly = annot_assembly
			entrezgene_mapping.gene = gene
			entrezgene_mapping.entrezgene_type = entrezgene_type
			
			session.add(entrezgene_mapping)
			param_obj.no_of_into_db += 1
		else:
			param_obj.no_of_entrezgene_mappings_already_in_db += 1
		"""
		for model_elem in element.findall("MODEL"):
			self.addGeneCommentary(model_elem, gene, session, param_obj=param_obj)
		
		param_obj.no_of_total += 1
		if param_obj.no_of_total%1000==0:
			try:
				session.flush()
				#session.expunge_all()
			except:
				sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
				import traceback
				traceback.print_exc()
				import pdb
				pdb.set_trace()
			
			report = getattr(param_obj, 'report', False)
			if report:
				sys.stderr.write('%s%s into db out of %s. %s gene(s) already_in_db.\n\
						%s entrezgene_mapping(s) already_in_db. %s gene2go(s) already in db.\n'%\
						('\x08'*100, param_obj.no_of_into_db, param_obj.no_of_total, param_obj.no_of_genes_already_in_db,\
						param_obj.no_of_entrezgene_mappings_already_in_db, param_obj.no_of_gene2go_already_in_db))
		
	def parse_xml_file(self, session, filename, tax_id=None, param_obj=None):
		"""
		2010-8-11
			
		"""
		counter = 0
		real_counter = 0
		if filename[-2:]=='gz':	#2010-8-12
			import gzip
			fHandler = gzip.open(filename)
		else:
			fHandler = open(filename)
		
		for event, elem in ElementTree.iterparse(fHandler):
			if elem.tag == 'ASSEMBLY':
				assembly_elem = elem
				chromosome = assembly_elem.get("CHROMOSOME")
				current_date = assembly_elem.get("CURRENT_DATE")
				
				start = int(assembly_elem.findtext('COORDSET/END5'))
				stop = int(assembly_elem.findtext('COORDSET/END3'))
				organism = assembly_elem.findtext("HEADER/ORGANISM")
				lineage = assembly_elem.findtext("HEADER/LINEAGE")
				date_last_touched = assembly_elem.findtext("HEADER/SEQ_LAST_TOUCHED/DATE")
				clone_name = assembly_elem.findtext("HEADER/CLONE_NAME")
				comment = 'CURRENT_DATE %s, SEQ_LAST_TOUCHED %s, lineage %s, %s'%(current_date, date_last_touched, lineage, organism)
				
				annot_assembly = AnnotAssembly.query.filter_by(acc_ver=clone_name).first()
				if not annot_assembly:
					annot_assembly = AnnotAssembly(acc_ver=clone_name, chromosome=chromosome, start=start, stop=stop, tax_id=tax_id, \
												sequence_type_id=1, comment=comment)
					session.add(annot_assembly)
					session.flush()
				
				protein_elements = elem.find('GENE_LIST/PROTEIN_CODING')
				RNA_elements = elem.find('GENE_LIST/RNA_GENES')
				TE_elements = elem.find('GENE_LIST/TRANSPOSABLE_ELEMENT_GENES')
				
				for protein_element in protein_elements:
					self.addGene(session, annot_assembly, protein_element, type_of_gene='protein-coding',\
								param_obj=param_obj)
					real_counter += 1
				for element in RNA_elements:
					self.addGene(session, annot_assembly, element, param_obj=param_obj)
					real_counter += 1
				for element in TE_elements:
					self.addGene(session, annot_assembly, element, param_obj=param_obj)
					real_counter += 1
				
				#repeat_elements = elem.find("REPEAT_LIST")
				#misc_elements = elem.find("MISC_INFO")
				
				#release memory
				elem.clear()
				counter += 1
			if self.report and counter%2000==0 and counter>0:
				sys.stderr.write("%s\t%s/%s"%('\x08'*20, counter, real_counter))
		sys.stderr.write('%s%s into db out of %s. %s gene(s) already_in_db.\n \
					%s entrezgene_mapping(s) already_in_db. %s gene2go(s) already in db.\n'%\
					('\x08'*100, param_obj.no_of_into_db, param_obj.no_of_total, param_obj.no_of_genes_already_in_db,\
					param_obj.no_of_entrezgene_mappings_already_in_db, param_obj.no_of_gene2go_already_in_db))
	
	def run(self):
		"""
		11-13-05 
			--db_connect()
			--parse_entrezgene_xml_file()
				--is_gi_valid_in_annot_assembly_table()
				--find_info_dict()
					--return_location_list()
				--submit_to_entrezgene_mapping_table()
		"""
		if self.debug:
			import pdb
			pdb.set_trace()
		
		sys.stderr.write("\tTotally, %d files to be processed.\n"%len(self.inputfiles))
		db = GenomeDatabase(drivername=self.drivername, username=self.db_user,
						password=self.db_passwd, hostname=self.hostname, database=self.dbname, schema=self.schema)
		db.setup(create_tables=False)	#2010-6-22
		session = db.session
		param_obj = PassingData(session=db.session, no_of_genes_already_in_db=0, no_of_entrezgene_mappings_already_in_db=0,\
					no_of_total=0, no_of_into_db=0, report=self.report, no_of_commentaries_already_in_db=0,\
					no_of_gene_segments_already_in_db=0, no_of_gene2go_already_in_db=0)
		for f in self.inputfiles:
			sys.stderr.write("%d/%d:\t%s\n"%(self.inputfiles.index(f)+1,len(self.inputfiles),f))
			self.parse_xml_file(session, f, tax_id=self.tax_id, param_obj=param_obj)
		
		session.flush()
		if self.commit:
			session.commit()
		else:
			session.rollback()

if __name__ == '__main__':
	#import pdb
	#pdb.set_trace()
	#from palos import ProcessOptions
	main_class = TAIRGeneXML2GenomeDB
	instance = TAIRGeneXML2GenomeDB(sys.argv)
	#po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
	#instance = main_class(**po.long_option2value)
	instance.run()



#!/usr/bin/env python3
"""
Description:
    A matrix stored in HDF5.
        This HDF5 file is composed of groups. At least one group, group0.
i.e.
    reader = HDF5MatrixFile(inputFname=filename, mode='r')
    reader = HDF5MatrixFile(filename, mode='r')
    for row in reader:
        ...
    tableObject = reader.getTableObject(tableName=tableName)
    for row in tableObject:
        ...
    
    rowDefinition = [('locus_id','i8'),('chr', HDF5MatrixFile.varLenStrType), ('start','i8'), ('stop', 'i8'), \
                ('score', 'f8'), ('MAC', 'i8'), ('MAF', 'f8')]
    headerList = [row[0] for row in rowDefinition]
    dtype = numpy.dtype(rowDefinition)
    
    writer = HDF5MatrixFile(inputFname=filename, mode='w', dtype=dtype)
    writer = HDF5MatrixFile(filename, mode='w', dtype=dtype)
    
    if writer:
        tableObject = writer.createNewTable(tableName=tableName, dtype=dtype)
        tableObject.setColIDList(headerList)
    elif outputFname:
        writer = HDF5MatrixFile(outputFname, mode='w', dtype=dtype, tableName=tableName)
        writer.writeHeader(headerList)
        tableObject = writer.getTableObject(tableName=tableName)
    cellList = []
    for data_obj in self.data_obj_ls:
        dataTuple = self._extractOutputRowFromDataObject(data_obj=data_obj)
        cellList.append(dataTuple)
    tableObject.writeCellList(cellList)
    if closeFile:
        writer.close()
        del writer
    
    
    #each number below is counting bytes, not bits
    rowDefinition = [('locus_id','i8'),('chromosome', HDF5MatrixFile.varLenStrType),
        ('start','i8'), ('stop', 'i8'), \
        ('score', 'f8'), ('MAC', 'i8'), ('MAF', 'f8')]
    if writer is None and filename:
        writer = HDF5MatrixFile(filename, mode='w', rowDefinition=rowDefinition, tableName=tableName)
        tableObject = writer.getTableObject(tableName=tableName)
    elif writer:
        tableObject = writer.createNewTable(tableName=tableName, rowDefinition=rowDefinition)
"""

import sys, os, math
import csv
import h5py
import numpy
import logging
from palos.utils import PassingData, PassingDataList,getListOutOfStr 
from palos.ProcessOptions import  ProcessOptions
from palos.io.MatrixFile import MatrixFile
varLenStrType = h5py.vlen_dtype(str)

class YHTableInHDF5Group(object):
    option_default_dict = {
        ('h5Group', 0, ): [None, '', 1, "the h5py group object."],\
        ('newGroup', 1, int): [0, '', 1, 
            "whether this is a new group or an existing group in a file"],\
        ('dataMatrixDtype', 0, ): ['f', '', 1, 
            'data type in the dataMatrix. candidates are i, f8, compound type, etc.'],\
        ('compression', 0, ): [None, '', 1, 
            'the compression engine for all underlying datasets'],\
        ('compression_opts', 0, ): [None, '', 1,
            'option for the compression engine, gzip level, or tuple for szip'],\
        
        }
    def __init__(self, **keywords):
        """
        dataMatrixDtype could be a compound type:
            http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html
            http://docs.scipy.org/doc/numpy/reference/generated/numpy.dtype.html
                
        #A record data type containing a 16-character string (in field name)
            #and a sub-array of two 64-bit floating-point number (in field grades):
        dt = numpy.dtype([('name', numpy.str_, 16), ('grades', numpy.float64, (2,))])
        
        my_dtype = numpy.dtype([('field1', 'i'), ('field2', 'f'), ('field3', varLenStrType)])
        
        #Using array-protocol type strings:
        #each number below is counting bytes, not bits
        >>> numpy.dtype([('a','f8'),('b','S10')])
        dtype([('a', '<f8'), ('b', '|S10')])
        
        #Using tuples. int is a fixed type, 3 the field's shape. void is a flexible type, here of size 10:
        numpy.dtype([('hello',(numpy.int,3)),('world',numpy.void,10)])
        
        #Using dictionaries. Two fields named 'gender' and 'age':
        numpy.dtype({'names':['gender','age'], 'formats':['S1',numpy.uint8]})
        
        #Offsets in bytes, here 0 and 25:
        numpy.dtype({'surname':('S25',0),'age':(numpy.uint8,25)})
        """
        self.ad = ProcessOptions.process_function_arguments(keywords,
            self.option_default_dict, error_doc=self.__doc__,
            class_to_have_attr=self)
        
        self.dataMatrixDSName = "dataMatrix"
        self.rowIDListDSName = "rowIDList"
        self.colIDListDSName = "colIDList"
        if not self.newGroup:
            self._readInData()
        else:
            self._createDatasetSkeletonForOneGroup(h5Group=self.h5Group,
                dtype=self.dataMatrixDtype)
        
        self.newWrite = True
        #a flag used to control whether it's first time to write stuff (first time=set whole matrix)
        self.rowIndexCursor = 0
        
    def _createDatasetSkeletonForOneGroup(self, h5Group=None, dtype=None):
        """
        """
        if h5Group is None:
            h5Group = self.h5Group
        if dtype is None:
            dtype = self.dataMatrixDtype
        self.dataMatrix = h5Group.create_dataset(self.dataMatrixDSName, shape=(1,),
            dtype=dtype, maxshape=(None, ),
            compression=self.compression, compression_opts=self.compression_opts)
        #by default it contains one "null" data point.
        self.dataMatrix.resize((0,))
        self.rowIDList = h5Group.create_dataset(self.rowIDListDSName, shape=(1,),
            dtype=varLenStrType, maxshape=(None,),
            compression=self.compression, compression_opts=self.compression_opts)
        self.rowIDList.resize((0,))
        self.colIDList = h5Group.create_dataset(self.colIDListDSName, shape=(1,),
            dtype=varLenStrType, maxshape=(None,),\
            compression=self.compression, compression_opts=self.compression_opts)
        self.colIDList.resize((0,))
        self.rowID2rowIndex = {}
        self.colID2colIndex = {}
    
    def _readInData(self):
        """
        """
        self.dataMatrix = self.h5Group[self.dataMatrixDSName]
        self.rowIDList = self.h5Group[self.rowIDListDSName]
        self.colIDList = self.h5Group[self.colIDListDSName]
        self._processRowIDColID()
        #pass the HDF5Group attributes to this object itself,
        #  it ran into "can't set attribute error".
        # conflict with existing property
        #for attributeName, attributeValue in self.h5Group.attrs.items():
        #	object.__setattr__(self, attributeName, attributeValue)
    
    def _processRowIDColID(self):
        """
        Similar to SNPData.processRowIDColID()
        """
        rowIDList = self.rowIDList
        colIDList = self.colIDList
        rowID2rowIndex = {}
        if rowIDList:
            for i in range(len(rowIDList)):
                rowID = rowIDList[i]
                rowID2rowIndex[rowID] = i
        
        colID2colIndex = {}
        if colIDList:
            for i in range(len(colIDList)):
                colID = colIDList[i]
                colID2colIndex[colID] = i
        
        self.rowID2rowIndex = rowID2rowIndex
        self.colID2colIndex = colID2colIndex
    
    def extendDataMatrix(self, dataMatrix=None, **keywords):
        """
        dataMatrix has to be 2D
        """
        if dataMatrix:
            m = len(dataMatrix)
            s = self.dataMatrix.shape[0]
            if self.newWrite:
                #defaultData in self.dataMatrix is of shape (1,)
                self.dataMatrix.resize((m,))
                self.dataMatrix[0:m] = dataMatrix
            else:
                self.dataMatrix.resize((s+m,))
                self.dataMatrix[s:s+m] = dataMatrix
            self.newWrite = False
    
    def writeCellList(self, cellList=None, **keywords):
        """
        call self.extendDataMatrix()
        """
        self.extendDataMatrix(dataMatrix=cellList, **keywords)
    
    def _appendID(self, idDataset=None, idValue=None):
        """
        """
        self._extendIDList(idDataset=idDataset, idList=[idValue])

    def _extendIDList(self, idDataset=None, idList=None):
        """
        """
        if idDataset is None:
            idDataset = self.colIDList
        
        if idList:
            m = len(idList)
            s = idDataset.shape[0]
            idDataset.resize((s+m,))
            idDataset[s:s+m] = idList

    def _setIDList(self, idDataset=None, idList=None):
        """
        """
        if idDataset is None:
            idDataset = self.colIDList
        
        if idList:
            m = len(idList)
            s = idDataset.shape[0]
            idDataset.resize((m,))
            idDataset[:] = idList
        
    def setColIDList(self, colIDList=None):
        """
        """
        self._setIDList(idDataset=self.colIDList, idList=colIDList)
        
    def setRowIDList(self, rowIDList=None):
        """
        """
        self._setIDList(idDataset=self.rowIDList, idList=rowIDList)
    
    def getColIndex(self, colID=None):
        """
        
        """
        colID2colIndex = self.colID2colIndex
        colIndex = None
        if colID2colIndex:
            colIndex = colID2colIndex.get(colID, None)
        return colIndex
    
    def getRowIndex(self, rowID=None):
        """
        """
        rowID2rowIndex = self.rowID2rowIndex
        rowIndex = None
        if rowID2rowIndex:
            rowIndex = rowID2rowIndex.get(rowID, None)
        return rowIndex
    
    def getCellDataGivenRowColID(self, rowID=None, colID=None):
        """
        """
        
        rowIndex = self.getRowIndex(rowID)
        colIndex = self.getColIndex(colID)
        
        cellData = None
        if rowIndex is not None and colIndex is not None:
            cellData = self.dataMatrix[rowIndex][colIndex]
        return cellData
    
    @property
    def name(self):
        return self.h5Group.name
    
    def __iter__(self):
        return self
    
    def next(self):
        """
        part of faking as a file object
        """
        pdata = PassingDataList()
        if self.rowIndexCursor<self.dataMatrix.shape[0]:
            row = self.dataMatrix[self.rowIndexCursor]
            for colID in self.colIDList:
                #iterate over colIDList is in the same order as the ascending order of colIndex
                #but iteration over self.colID2colIndex is not in the same
                #   order as the ascending order of colIndex
                colIndex = self.colID2colIndex.get(colID)
                setattr(pdata, colID, row[colIndex])
            self.rowIndexCursor += 1
        else:
            raise StopIteration
        return pdata
    
    def reset(self):
        """
        part of faking as a file object
        """
        self.rowIndexCursor = 0
    
    def constructColName2IndexFromHeader(self):
        """
        added so that YHSingleTableFile could be used as HDF5MatrixFile
        """
        return self.colID2colIndex
    
    
    def getColIndexGivenColHeader(self, colHeader=None):
        """
        this is from the combined column header list.
        """
        return self.colID2colIndex.get(colHeader)
    
    def addAttribute(self, name=None, value=None, overwrite=True):
        """
        
        """
        if name in self.h5Group.attrs:
            logging.warn("h5Group %s already has attribute %s=%s."%(
                self.name, name, value))
            if overwrite:
                self.h5Group.attrs[name] = value
            else:
                return False
        else:
            self.h5Group.attrs[name] = value
        #pass the HDF5Group attributes to this object itself,
        #  it ran into "can't set attribute error".
        #  conflict with existing property
        #object.__setattr__(self, name, value)
        #setattr(self, name, value)
        return True
    
    def addAttributeDict(self, attributeDict=None):
        """
        """
        addAttributeDictToYHTableInHDF5Group(tableObject=self, attributeDict=attributeDict)
    
    def getAttribute(self, name=None, defaultValue=None):
        return self.h5Group.attrs.get(name, defaultValue)
    
    def getAttributes(self):
        return self.h5Group.attrs
    
    def getListAttributeInStr(self, name=None):
        """
        this attribute must be a list or array
        """
        attr_in_str = ''
        attributeValue = self.getAttribute(name=name)
        if attributeValue is not None and type(attributeValue)==numpy.ndarray:
            if hasattr(attributeValue, '__len__') and attributeValue.size>0:
                ls = map(str, attributeValue)
                attr_in_str = ','.join(ls)
        return attr_in_str
    
    def addObjectAttributeToSet(self, attributeName=None, setVariable=None):
        """
        do not add an attribute to the set if it's not available or if it's none
        """
        attributeValue = self.getAttribute(attributeName, None)
        if attributeValue is not None and setVariable is not None:
            setVariable.add(attributeValue)
        return setVariable
    
    def addObjectListAttributeToSet(self, attributeName=None,
        setVariable=None, data_type=None):
        """
        """
        attributeValue = self.getAttribute(attributeName, None)
        flag = False
        if type(attributeValue)==numpy.ndarray:
            #"if attributeValue" fails for numpy array
            if hasattr(attributeValue, '__len__') and attributeValue.size>0:
                flag = True
        elif attributeValue or attributeValue == 0:
            flag = True
        if flag and setVariable is not None:
            if type(attributeValue)==str:
                attributeValueList = getListOutOfStr(attributeValue, data_type=data_type, separator1=',', separator2='-')
            else:
                attributeValueList = attributeValue
            setVariable |= set(list(attributeValueList))
        return setVariable
    
    def close(self):
        """
        deprecated. dimension set to 0 upon creation.
        resize the data matrix to zero if it's in write mode and self.newWrite is still True
        """
        if self.newGroup and self.newWrite:
            #in write mode and no writing has happened.
            self.dataMatrix.resize((0,))
            #or you can delete it
            #del self.h5Group[self.dataMatrixDSName]
            #self.h5Group.delete_dataset( ...
    

class HDF5MatrixFile(MatrixFile):
    varLenStrType = varLenStrType
    #convenient for outside program to access this variable length string type
    __doc__ = __doc__
    option_default_dict = MatrixFile.option_default_dict.copy()
    option_default_dict.update({
        ('tableName', 0, ): [None, '', 1, 
            "name for the first table, default is $tableNamePrefix\0. "],\
        ('tableNamePrefix', 0, ): ['table', '', 1, 
            "prefix for all table's names"],\
        ('rowDefinition', 0, ): [None, '', 1,
            "data type list for a compound dtype. It overwrites dtype. i.e. a list of i.e. ('start','i8')"],\
        ('dtype', 0, ): [None, '', 1, 
            'data type in the first table to be created. candidates are i, f8, etc.'],\
        ('compression', 0, ): ['gzip', '', 1,
            'the compression engine for all underlying datasets, gzip, szip, lzf'],\
        ('compression_opts', 0, ): [None, '', 1, 
            'option for the compression engine, gzip level, or tuple for szip'],\
        })
    def __init__(self, inputFname=None, **keywords):
        self.ad = ProcessOptions.process_function_arguments(keywords, 
            self.option_default_dict, error_doc=self.__doc__, \
            class_to_have_attr=self)
        if not self.inputFname:
            self.inputFname = inputFname
        
        self.header = None
        self.combinedColIDList = None	#same as header
        self.combinedColID2ColIndex = None
        
        self.hdf5File = h5py.File(self.inputFname, self.mode)
        self.tableObjectList = []
        self.tablePath2Index = {}
        
        if self.mode=='r':
            self._readInData()
        elif self.mode=='w':
            self.createNewTable(tableName=self.tableName, dtype=self.dtype, rowDefinition=self.rowDefinition)
        
        self.rowIndexCursor = 0	#2012.11.16 for iteration
    
    def createNewTable(self, tableName=None, dtype=None, rowDefinition=None):
        """
        2012.11.20 add argument rowDefinition
        2012.11.19
        """
        colIDList = None
        if rowDefinition:
            colIDList = [row[0] for row in rowDefinition]
            dtype = numpy.dtype(rowDefinition)
        if dtype is None:
            dtype = self.dtype
        if tableName is None:
            tableName = self._getNewTableName()
        tableObject = YHTableInHDF5Group(
            h5Group=self.hdf5File.create_group(tableName),
            newGroup=True, dataMatrixDtype=dtype,
            compression=self.compression,
            compression_opts=self.compression_opts)
        if colIDList:
            tableObject.setColIDList(colIDList)
        self._appendNewTable(tableObject)
        return tableObject
    
    def _appendNewTable(self, tableObject=None):
        if tableObject:
            tableName = tableObject.name
            ##the tableName preceds with "/"
            if tableName in self.tablePath2Index:	 
                logging.error("table %s already in self.tablePath2Index, index=%s."%(
                    tableName,
                    self.tablePath2Index.get(tableName)))
                sys.exit(3)
            self.tablePath2Index[tableName] = len(self.tablePath2Index)
            self.tableObjectList.append(tableObject)
    
    def _getNewTableName(self, tableNamePrefix=None):
        """
        2012.11.19
        """
        if not tableNamePrefix:
            tableNamePrefix = self.tableNamePrefix
        i = len(self.tableObjectList)
        tableName = "%s%s"%(tableNamePrefix, i)
        while "/"+tableName in self.tablePath2Index:
            #stop until a unique name shows up
            i += 1
            tableName = "%s%s"%(tableNamePrefix, i)
        return tableName
    
    def _readInData(self):
        """
        """
        for tableName, h5Group in self.hdf5File.items():
            tableObject = YHTableInHDF5Group(h5Group=h5Group, newGroup=False,\
                                        dataMatrixDtype=self.dtype)
            self._appendNewTable(tableObject)
        self._setupCombinedColumnIDMapper()
        
    def _setupCombinedColumnIDMapper(self,):
        """
        """
        self.combinedColID2ColIndex = {}
        self.header = []
        self.combinedColIDList = self.header
        for tableObject in self.tableObjectList:
            colIDList = tableObject.colIDList
            for colID in colIDList:
                if colID in self.combinedColID2ColIndex:
                    sys.stderr.write("Error: column ID %s already used in column %s.\n"%(colID, self.combinedColID2ColIndex.get(colID)))
                    sys.exit(3)
                else:
                    self.combinedColID2ColIndex[colID] = len(self.combinedColID2ColIndex)
                    self.header.append(colID)
    
    def constructColName2IndexFromHeader(self):
        """
        overwrite parent function
        """
        return self.combinedColID2ColIndex
    
    def getColIndexGivenColHeader(self, colHeader=None):
        """
        this is from the combined column header list.
        """
        return self.combinedColID2ColIndex.get(colHeader)
    
    def getTableObject(self, tableIndex=None, tableName=None):
        """
        """
        tableObject = None
        if tableIndex is not None:
            tableObject = self.tableObjectList[tableIndex]
        elif tableName:
            if tableName[0]!='/':
                #bugfix, add root in front if not there.
                tableName = '/'+tableName
            tableIndex = self.tablePath2Index.get(tableName)
            if tableIndex is not None:
                tableObject = self.tableObjectList[tableIndex]
        else:	#return first table
            tableObject = self.tableObjectList[0]
        return tableObject
    
    def getColIndex(self, colID=None, tableIndex=None, tableName=None):
        """
        
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        return tableObject.getColIndex(colID=colID)
    
    def getRowIndex(self, rowID=None, tableIndex=None, tableName=None):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        return tableObject.getRowIndex(rowID=rowID)
    
    def getCellDataGivenRowColID(self, rowID=None, colID=None, tableIndex=None, tableName=None):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        return tableObject.getCellDataGivenRowColID(rowID=rowID, colID=colID)
    
    def __iter__(self):
        return self
    
    def next(self):
        """
        combine the numeric and str row (if both exist). and return the combined row
        """
        
        row = None
        pdata = PassingDataList()
        for tableObject in self.tableObjectList:
            if self.rowIndexCursor<tableObject.dataMatrix.shape[0]:
                if row is None:
                    row = list(tableObject.dataMatrix[self.rowIndexCursor])
                else:
                    rowAppend = tableObject.dataMatrix[self.rowIndexCursor]
                    row += list(rowAppend)
            else:
                raise StopIteration
            for colID  in self.header:	#iteration over header is in the same order as the ascending order of colIndex
                #but iteration over self.colID2colIndex is not in the same order as the ascending order of colIndex
                colIndex = self.combinedColID2ColIndex.get(colID)
                if colIndex < len(row):
                    setattr(pdata, colID, row[colIndex])
        self.rowIndexCursor += 1
        return pdata
    
    def close(self):
        """
        #2012.12.4 close the table objects first
        """
        self.hdf5File.close()
        del self.hdf5File
    
    def writeHeader(self, headerList=None, tableIndex=None, tableName=None):
        """
        only the first table
        """
        self.setColIDList(colIDList=headerList, tableIndex=tableIndex,
            tableName=tableName)
    
    def writeOneCell(self, oneCell=None, tableIndex=None, tableName=None):
        """
        mimic csv's writerow()
        each cell must be a tuple (readable only). list is not acceptable. 
        it's not very efficient as it's resizing the dataMatrix all the time.
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        tableObject.extendDataMatrix(dataMatrix=[oneCell])

    def writeCellList(self, cellList=None, tableIndex=None, tableName=None):
        """
        for bulk writing. more efficient.
        each cell must be a tuple (readable only). list is not acceptable. 
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        tableObject.extendDataMatrix(dataMatrix=cellList)
    
    
    def setColIDList(self, colIDList=None, tableIndex=None, tableName=None):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        tableObject.setColIDList(colIDList=colIDList)
        
    def setRowIDList(self, rowIDList=None, tableIndex=None, tableName=None):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        tableObject.setRowIDList(rowIDList=rowIDList)
    
    def addAttribute(self, name=None, value=None, overwrite=True,
        tableIndex=None, tableName=None):
        """
        find the tableObject and let it do the job
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        return tableObject.addAttribute(name=name, value=value, overwrite=overwrite)
    
    def addAttributeDict(self, attributeDict=None, tableObject=None):
        if tableObject is None:
            tableObject = self.getTableObject()
        
        addAttributeDictToYHTableInHDF5Group(tableObject=tableObject, attributeDict=attributeDict)
    
    def getAttribute(self, name=None, defaultValue=None, tableIndex=None, tableName=None):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        return tableObject.getAttribute(name=name, defaultValue=defaultValue)
    
    def getAttributes(self, tableIndex=None, tableName=None):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex,
            tableName=tableName)
        return tableObject.getAttributes()
    
    def getListAttributeInStr(self, name=None, tableIndex=None, tableName=None):
        """
        this attribute must be a list or array
        """
        tableObject = self.getTableObject(tableIndex=tableIndex,
            tableName=tableName)
        return tableObject.getListAttributeInStr(name=name)
    
    def run(self):
        """
        """
        
        if self.debug:
            import pdb
            pdb.set_trace()

def addAttributeDictToYHTableInHDF5Group(tableObject=None, attributeDict=None):
    """
    convenient function
        attributeValue could not be high-level python objects, such as list, set.
        numpy.array could replace list.
    """
    if tableObject is not None and attributeDict is not None:
        for attributeName, attributeValue in attributeDict.items():
            doItOrNot = False
            if type(attributeValue)==numpy.ndarray:
                if hasattr(attributeValue, '__len__') and attributeValue.size>0:
                    doItOrNot = True
            elif attributeValue or attributeValue == 0:
                #empty array will be ignored but not 0
                doItOrNot = True
            if doItOrNot:
                tableObject.addAttribute(name=attributeName, value=attributeValue)
    

if __name__ == '__main__':
    main_class = HDF5MatrixFile
    po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    instance.run()



#!/usr/bin/env python3
"""
2009-11-2
    module for various algorithms
"""
import os, sys
from palos.ProcessOptions import  ProcessOptions
from palos.utils import PassingData
import copy
import numpy

def listSubsets(element_ls, subset_size=None):
    """
    Example:
        element_ls = [1,6,7]
        listSubsets(element_ls, subset_size=1)
    
        listSubsets(element_ls, subset_size=2)
    
        listSubsets(element_ls)
    
    2009-11-2
        list all possible subsets of element_ls. If subset_size is given, only produce subsets of that size.
            Otherwise, all of them.
        It's backtracking algorithm at play.
    """
    sys.stderr.write("Listing all subsets of a list of %s elements ... "%(len(element_ls)))
    no_of_elements = len(element_ls)
    candidate_ls_ls = [[0,1]]	# 0 or 1 in k-th candidate_ls_ls signifies whether element_ls[k] would be included or not.
    #for i in range(no_of_elements):
    #	candidate_ls_ls.append([0, 1])
    
    one_solution=[]
    solution_ls=[]
    k = 0
    while k>=0:
        while len(candidate_ls_ls[k])>0:
            next_element = candidate_ls_ls[k].pop()
            if len(one_solution)==no_of_elements:
                one_solution[k] = next_element
            else:
                one_solution.append(next_element)
            k += 1
            if k==no_of_elements:
                if subset_size is None or sum(one_solution)==subset_size:
                    one_subset = []
                    for i in range(no_of_elements):
                        if one_solution[i]==1:
                            one_subset.append(element_ls[i])
                    solution_ls.append(one_subset)		# python data by reference. have to use copy to sever the tie.
                break
            if len(candidate_ls_ls)<=k:	# still growing
                candidate_ls_ls.append([0,1])
            else:	# fully grown, now just replace the candidate list
                candidate_ls_ls[k] = [0, 1]
        k -= 1
    sys.stderr.write("Done.\n")
    return solution_ls

def simpleCandidateSetGenerator(k, element_ls=range(2), max_solution_size=5):
    """
    2010-4-21
        candidates will be from element_ls as far as k is less than max_solution_size.
        used as argument candidateSetGenerator by enumerateCombinations().
    """
    if k<max_solution_size:
        return element_ls[:]	#[:] is a must. otherwise, range(2) is used by reference.
    else:
        return []


def enumerateCombinations(candidateSetGenerator=None):
    """
    Examples:
        # 2010-4-21 directly use simpleCandidateSetGenerator
        enumerateCombinations(simpleCandidateSetGenerator)
        
        # 2010-4-21 modify the max_solution_size to be 4
        f = lambda x: simpleCandidateSetGenerator(x, element_ls=range(2), max_solution_size=4)
        enumerateCombinations(f)
    
    2010-4-21
        This function enumerates combinations given candidate sets at each position.
        The backtrack algorithm will stop when the candidateSetGenerator returns nothing.
        candidateSetGenerator accepts argument k, index of the element in one solution.
    """
    sys.stderr.write("Enumerating combinations of candidates ...")
    candidate_set_ls = [candidateSetGenerator(0)]	# first candidate
    
    one_solution=[]
    solution_ls = []
    k = 0
    while k>=0:
        while len(candidate_set_ls[k])>0:
            next_element = candidate_set_ls[k].pop()
            if len(one_solution)>k:
                one_solution[k] = next_element
            else:
                one_solution.append(next_element)
            k += 1
            candidate_set = candidateSetGenerator(k)
            if not candidate_set:	# no more candidate set, this is a solution.
                solution_ls.append(one_solution[:])	# [:] is a must. otherwise, Every single item in solution_ls will be same.
                # because each one is a reference to one_solution. 
            if len(candidate_set_ls)<=k:	# still growing
                candidate_set_ls.append(candidate_set)
            else:	# fully grown, now just replace the candidate list
                candidate_set_ls[k] = candidate_set
        k -= 1
    sys.stderr.write("Done.\n")
    return solution_ls

def ltsFit(x_ls, y_ls, fractionUsed=0.6, startX=1, stopX=5):
    """
    2010-6-1
        solve the computing node hang-up (I/O stuck) issue by adding these:
            import ROOT
            try:	# 2010-5-31 old version (5.18.0) doesn't have IgnoreCommandLineOptions.
                ROOT.PyConfig.IgnoreCommandLineOptions = True	#otherwise
                # Warning in <TApplication::GetOptions>: file <output file by -o > has size 0, skipping
            except:
                pass
            try:	# 2010-5-31  disable .StartGuiThread
                ROOT.PyConfig.StartGuiThread = 0
            except:
                pass
    2010-5-30
        return chiSquare as well
    2010-5-21
        use ROOT to do least trimmed square (LTS) fitting:
            fit the y=a+bx with trimming fraction = 1-fractionUsed.
    
    Example:
    
    import numpy
    x_ls = numpy.array(range(100), numpy.float)
    y_ls = x_ls/2.
    for i in range(len(y_ls)):
        import random
        new_y = random.random()-0.5
        y_ls[i] += new_y
    
    # mess up some portion of y
    for i in range(5):
        import random
        new_y = random.random()
        new_y_index = random.sample(range(100),1)
        y_ls[new_y_index[0]] = new_y
    import numpy
    x_ls = numpy.array([ 2.64884758,  3.51235008,  2.83090925,  3.41229248,  3.01451969,\
    2.49899888,  3.69988108,  2.74896216,  3.05307841,  3.75705409,\
    3.08653784,  3.10703993,  3.61071348,  3.21285319,  2.91460752,\
    3.53737831,  3.06333303,  3.35391617,  3.43568516,  3.34429312,\
    3.31576061,  2.8007164 ,  2.73639655,  3.14690256,  3.10174704,\
    2.80888581,  2.72754121,  2.90064001,  3.19270658,  3.50596333,\
    2.61804676,  3.18127131,  3.27542663,  3.09586573], dtype=numpy.float32)	# numpy.float32 is not supported by ROOT
    y_ls = numpy.array([ 2.52827311,  3.27265358,  2.36172366,  2.95760489,  2.50920248,\
    2.3443923 ,  3.23502254,  2.35410833,  2.50582743,  2.48501062,\
    2.82510138,  2.70799541,  2.43136382,  2.76342535,  2.45178652,\
    3.08224201,  2.26481771,  2.7387805 ,  3.23274207,  2.82769203,\
    2.25042009,  2.56702638,  2.4082365 ,  2.44793224,  2.65127802,\
    2.57460976,  2.43136382,  2.39005065,  2.70027065,  3.04452848,\
    2.28555727,  2.71933126,  2.6468935 ,  2.54157925], dtype=numpy.float32)
    
    fit_y_ls = ltsFit(x_ls, y_ls)
    
    import pylab
    pylab.plot(x_ls, y_ls, '.')
    pylab.plot(x_ls, fit_y_ls, '.')
    pylab.legend(['raw data','fitted'])
    pylab.show()
    sys.exit(0)
    
    """
    import ROOT
    try:	# 2010-5-31 old version (5.18.0) doesn't have IgnoreCommandLineOptions.
        ROOT.PyConfig.IgnoreCommandLineOptions = True	#otherwise
        # Warning in <TApplication::GetOptions>: file <output file by -o > has size 0, skipping
    except:
        pass
    try:	# 2010-5-31  disable .StartGuiThread
        ROOT.PyConfig.StartGuiThread = 0
    except:
        pass
    
    #ROOT.gROOT.Reset()	# 2010-5-31 dont' know what this is  for.
    ROOT.gROOT.SetBatch(True)	#to avoid interative mode (drawing canvas and etc.)	
    from ROOT import TFormula, TF1, TGraph
    import numpy
    lm = TF1('lm', 'pol1', startX, stopX)	#[0]+[1]*x is essentially same as pol1 but option rob in Fit() only works with pol1.
    #ROOT is very dtype-sensitive. numpy.float32 won't work.
    if hasattr(x_ls, 'dtype') and x_ls.dtype==numpy.float:
        pass
    else:
        sys.stderr.write('converting x_ls')
        x_ls = numpy.array(x_ls, dtype=numpy.float)
        sys.stderr.write(".\n")
    if hasattr(y_ls, 'dtype') and y_ls.dtype==numpy.float:
        pass
    else:
        sys.stderr.write('converting y_ls')
        y_ls = numpy.array(y_ls, dtype=numpy.float)
        sys.stderr.write(".\n")
    gr = TGraph(len(x_ls), x_ls, y_ls)
    gr.Fit(lm, "+rob=%s"%fractionUsed)
    fit = gr.GetFunction('lm')
    chiSquare = fit.GetChisquare();
    fit_y_ls = []
    for x in x_ls:
        fit_y_ls.append(fit.Eval(x))
    from utils import PassingData
    return PassingData(fit_y_ls=fit_y_ls, chiSquare=chiSquare)

def smoothFullData(fullData, smooth_type_id=3, no_of_overview_points=1500):
    """
    2010-9-25
        fullData is a list of objects. Each object has attributes, chromosome, start, stop.
        this function reduces the fullData to the number of data points that is <= no_of_overview_points.
    """
    import numpy
    if smooth_type_id is not None:
        smooth_type_id = int(smooth_type_id)
    if no_of_overview_points is not None:
        no_of_overview_points = int(no_of_overview_points)
    if len(fullData)<=no_of_overview_points:
        overviewData = fullData
    else:
        xStart = fullData[0]['start']
        xStop = fullData[-1]['stop']
        chromosome = fullData[0]['chromosome']
        overviewData = []
        if smooth_type_id==1:	#get the top no_of_overview_points
            yStart_ls = [data['yStart'] for data in fullData]
            top_data_index_ls = numpy.argsort(yStart_ls)[-no_of_overview_points:]
            top_data_index_ls.sort()	#to preserve the original order in fullData
            for i in top_data_index_ls:
                overviewData.append(fullData[i])
        else:
            windowSize = (xStop-xStart + 1)/(no_of_overview_points-1)	#windows overlap, so that's why 2*
            windowStart = xStart
            windowStop = min(windowStart + windowSize, xStop)
            nextWindowStart = windowStart + int(windowSize/2)
            data_in_window = []
            no_of_data_before_nextWindowStart = 0
            for i in range(len(fullData)):
                data = fullData[i]
                if (data['start']>=windowStart and data['start']<=windowStop) or ( data['stop'] <=windowStop and data['stop']>=windowStart):
                    data_in_window.append(data)
                    if data['stop']<nextWindowStart:
                        no_of_data_before_nextWindowStart += 1
                elif data['start']>windowStop:
                    if smooth_type_id==3:
                        value = max([row['yStart'] for row in data_in_window])
                    elif smooth_type_id==4:
                        value = sum([row['yStart']*(max(0, min(row['stop'], windowStop) - max(row['start'], windowStart))+ 1) \
                                    for row in data_in_window])/windowSize
                    elif smooth_type_id==2:
                        value = numpy.median([row['yStart'] for row in data_in_window])
                    else:
                        #including smooth_type_id==1
                        value = numpy.median([row['yStart'] for row in data_in_window])
                    overviewData.append(dict(start=windowStart, stop=windowStop, yStart=value, chromosome=chromosome))
                    windowStart = nextWindowStart
                    windowStop = min(windowStart + windowSize, xStop)
                    nextWindowStart = windowStart + int(windowSize/2)
                    data_in_window = []
                    no_of_data_before_nextWindowStart = 0
                    data_in_window.append(data)
    return overviewData

def testSmoothFullData():
    fullData = []
    import random
    for i in range(5000):
        fullData.append({'start':i*1000+1, 'stop':(i+1)*1000, 'yStart':random.random()*10, 'chromosome':1 })
    overviewData = smoothFullData(fullData, smooth_type_id=3, no_of_overview_points=1500)

class LD(object):
    """
    2010-9-30
        moved from palos/SNP.py to calculate LD
    """
    def __init__(self):
        pass
    
    @classmethod
    def fill_in_snp_allele2index(cls, allele, allele2index):
        """
        2010-9-30
            
        2008-09-05
            used in calLD
        """
        if allele not in allele2index:
            allele2index[allele] = len(allele2index)
        return allele2index[allele]
    
    @classmethod
    def calLD(cls, locus1_allele_ls, locus2_allele_ls, locus1_id=None, locus2_id = None):
        """
        2010-9-30
            copied from palos/SNP.py.
            locus1_allele_ls, locus2_allele_ls should be bi-allelic.
            If locus1_allele_ls and locus2_allele_ls are of different size, the extra elements are discarded.
        2008-09-05
            adapted from variation.src.misc's LD.calculate_LD class
            only deal with 2-allele loci
            skip if either is NA, or if both are heterozygous (not phased)
        """
        counter_matrix = numpy.zeros([2,2])	#only 2 alleles
        snp1_allele2index = {}
        snp2_allele2index = {}
        no_of_individuals = min(len(locus1_allele_ls), len(locus2_allele_ls))
        for k in range(no_of_individuals):
            snp1_allele = locus1_allele_ls[k]
            snp2_allele = locus2_allele_ls[k]
            snp1_allele_index = cls.fill_in_snp_allele2index(snp1_allele, snp1_allele2index)
            snp2_allele_index = cls.fill_in_snp_allele2index(snp2_allele, snp2_allele2index)
            if snp1_allele_index>1 or snp2_allele_index>1:	#ignore the 3rd allele
                continue
            counter_matrix[snp1_allele_index, snp2_allele_index] += 1
            #counter_matrix[snp1_allele_index, snp2_allele_index] += 1	#this is to mimic the diploid.
        PA = sum(counter_matrix[0,:])
        Pa = sum(counter_matrix[1,:])
        PB = sum(counter_matrix[:,0])
        Pb = sum(counter_matrix[:,1])
        total_num = float(PA+Pa)
        try:
            PA = PA/total_num
            Pa = Pa/total_num
            PB = PB/total_num
            Pb = Pb/total_num
            PAB = counter_matrix[0,0]/total_num
            D = PAB-PA*PB
            PAPB = PA*PB
            PAPb = PA*Pb
            PaPB = Pa*PB
            PaPb = Pa*Pb
            Dmin = max(-PAPB, -PaPb)
            Dmax = min(PAPb, PaPB)
            if D<0:
                D_prime = D/Dmin
            else:
                D_prime = D/Dmax
            r2 = D*D/(PA*Pa*PB*Pb)
        except:	#2008-01-23 exceptions.ZeroDivisionError, Dmin or Dmax could be 0 if one of(-PAPB, -PaPb)  is >0 or <0
            sys.stderr.write('Unknown except, ignore: %s\n'%repr(sys.exc_info()[0]))
            return None
        allele_freq = (min(PA, Pa),min(PB, Pb))
        return_data = PassingData()
        return_data.D = D
        return_data.D_prime = D_prime
        return_data.r2 = r2
        return_data.allele_freq = allele_freq
        return_data.snp_pair_ls = (locus1_id, locus2_id)
        return_data.no_of_pairs = total_num
        return return_data

if __name__ == '__main__':
    import pdb
    pdb.set_trace()
    
    testSmoothFullData()
    
    element_ls = [1,6,7]
    print(listSubsets(element_ls, subset_size=1))
    print(listSubsets(element_ls, subset_size=2))
    print(listSubsets(element_ls))
